from .contrastive_loss import ContrastiveLoss
from .margin_loss import MarginLoss
from .triplet_loss import TripletLoss
from .angular_loss import AngularLoss
from .circle_loss import CircleLoss
from .fast_ap_loss import FastAPLoss
from .lifted_structure_loss import LiftedStructureLoss
from .multi_similarity_loss import MultiSimilarityLoss
from .n_pair_loss import NPairLoss
from .signal_to_noise_ratio_loss import SignalToNoiseRatioLoss
from .pos_pair_loss import PosPairLoss