# recorder protocol
TO_RECORD_LIST = 'to_record_list'
CSV_FOLDER = 'csv'
MODEL_FOLDER = 'model'
BOARD_FOLDER = 'board'
WANDB_FOLDER = 'wandb'
TO_SAVE_LIST = 'to_save_list'
DEVICE = 'device'
BEST_MODEL_SUFFIX = '_best.ckpt'
STEP_MODEL_SUFFIX = '_{}.ckpt'
LOAD_EXCEPT_FUNC = lambda x: x[7:]