"""
This module is designed for recording:

1. support ``Recorder``. Class ``WithRecorder`` allows to use ``recorder`` to save statistics during training using csv and tensorboard.

"""

from .with_recorder import WithRecorder