from .mlp import MLP
from .mlp import BatchNormMLP
from .identity import Identity
from .resnet50 import resnet50
from .deit import deit_small_patch16_224 as DeiT_S
# from pretrainedmodels import bninception
from .bninception import bninception
from torchvision.models import googlenet