from .default_collector import DefaultCollector
from .proxy_collector import ProxyCollector
from .moco_collector import MoCoCollector
from .simsiam_collector import SimSiamCollector
from .hdml_collector import HDMLCollector
from .daml_collector import DAMLCollector
from .dvml_collector import DVMLCollector